# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import rx
from dependency_injector import providers

from krules_core.arg_processors import BaseArgProcessor
from krules_core import RuleConst
from krules_core.base_functions import RuleFunctionBase, inspect
from krules_core.core import RuleFactory
from krules_core.providers import event_router_factory, subject_factory, proc_events_rx_factory

filters = RuleConst.FILTERS
processing = RuleConst.PROCESSING
rulename = RuleConst.RULENAME
processed = RuleConst.PROCESSED

@pytest.fixture
def router():
    router = event_router_factory()
    router.unregister_all()
    proc_events_rx_factory.override(providers.Singleton(rx.subjects.ReplaySubject))

    return event_router_factory()

def _assert(expr):
    assert expr
    return True

def test_simple_callable():

    class SimpleSet(RuleFunctionBase):

        def execute(self, arg1, arg2, arg3, arg4, arg5):

            self.payload["arg1"] = arg1
            self.payload["arg2"] = arg2
            self.payload["arg3"] = arg3
            self.payload["arg4"] = arg4
            self.payload["arg5"] = arg5

    RuleFactory.create(
        "test-simple-callable",
        subscribe_to="test-argprocessors-callables",
        data={
            processing: [
                SimpleSet(lambda: 1, 2, arg3=lambda: 3, arg4=4, arg5=lambda p: "I'll never be called")
            ]
        }
    )

    payload = {}
    event_router_factory().route("test-argprocessors-callables", "test-0", payload)

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-simple-callable" and _assert(
            x[processing][0]["args"][0] == 1
            and x[processing][0]["kwargs"]["arg3"] == 3
            and x[processing][0]["kwargs"]["arg4"] == 4
            and isinstance(x[processing][0]["kwargs"]["arg5"], str)
        )
    )

    assert payload["arg1"] == 1
    assert payload["arg2"] == 2
    assert payload["arg3"] == 3
    assert payload["arg4"] == 4
    assert inspect.isfunction(payload["arg5"])


def test_with_self():

    class WithSelfSet(RuleFunctionBase):

        def execute(self, arg1, arg2, arg3):

            self.payload["arg1"] = arg1
            self.payload["arg2"] = arg2
            self.payload["arg3"] = arg3

    RuleFactory.create(
        "test-with-self",
        subscribe_to="test-argprocessors-self",
        data={
            processing: [
                WithSelfSet(lambda self: self.payload["value_from"],
                            arg2=lambda self: self.subject.get("value_from"),
                            arg3=lambda p: "I'll never be called")
            ]
        }
    )

    payload = {"value_from": 1}
    subject = subject_factory("test-1")

    subject.set("value_from", 2)

    event_router_factory().route("test-argprocessors-self", subject, payload)
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-with-self" and _assert(
            x[processing][0]["args"][0] == 1
            and x[processing][0]["kwargs"]["arg2"] == 2
            and isinstance(x[processing][0]["kwargs"]["arg3"], str))
    )

    assert payload["arg1"] == 1
    assert payload["arg2"] == 2
    assert inspect.isfunction(payload["arg3"])


def test_with_payload_and_subject():

    class WithPayloadSet(RuleFunctionBase):

        def execute(self, arg1, arg2, arg3):

            self.payload["arg1"] = arg1
            self.payload["arg2"] = arg2
            self.payload["arg3"] = arg3

    RuleFactory.create(
        "test-with-payload-and-subject",
        subscribe_to="test-argprocessors-payload-and-subject",
        data={
            processing: [
                WithPayloadSet(lambda payload: payload["value_from"],
                               arg2=lambda subject: subject.get("value_from"),
                               arg3=lambda p: "I'll never be called")
            ]
        }
    )

    _payload = {"value_from": 1}
    _subject = subject_factory("test-1")

    _subject.set("value_from", 2)

    event_router_factory().route("test-argprocessors-payload-and-subject", _subject, _payload)

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-with-payload-and-subject" and _assert(
            x[processing][0]["args"][0] == 1 and x[processing][0]["kwargs"]["arg2"] == 2
            and isinstance(x[processing][0]["kwargs"]["arg3"], str))
    )

    assert _payload["arg1"] == 1
    assert _payload["arg2"] == 2
    assert inspect.isfunction(_payload["arg3"])


def test_extend_jp_match():

    import jsonpath_rw_ext as jp
    from krules_core.arg_processors import processors

    class JPMatchSet(RuleFunctionBase):

        def execute(self, values, arg2):

            self.payload["values"] = values
            self.payload["elem-2"] = arg2

    class JPPayloadMatchBase:

        def __init__(self, expr):
            self._expr = expr

        def match(self, instance):
            raise NotImplementedError()

    class jp_match(JPPayloadMatchBase):

        def match(self, instance):
            return jp.match(self._expr, instance.payload)

    class jp_match1(JPPayloadMatchBase):

        def match(self, instance):
            return jp.match1(self._expr, instance.payload)

    class JPProcessor(BaseArgProcessor):

        @staticmethod
        def interested_in(arg):
            return isinstance(arg, JPPayloadMatchBase)

        def process(self, instance):
            return self._arg.match(instance)

    processors.append(JPProcessor)

    RuleFactory.create(
        "test-with-jp-expr",
        subscribe_to="test-argprocessors-jp-match",
        data={
            processing: [
                JPMatchSet(jp_match("$.elems[*].value"),
                           jp_match1("$.elems[?id==2]"))
            ]
        }
    )

    payload = {
        "elems": [
            {
                "id": 1,
                "value": "a"
            },
            {
                "id": 2,
                "value": "b"
            }
        ]
    }

    event_router_factory().route("test-argprocessors-jp-match", "test-0", payload)

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-with-jp-expr" and _assert(
            x[processing][0]["args"][0] == ['a', 'b']
            and x[processing][0]["args"][1] == {"id": 2, "value": "b"})
    )

    assert payload["values"] == ['a', 'b']
    assert payload["elem-2"]["id"] == 2 and payload["elem-2"]["value"] == "b"


# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import json

class PayloadConst(object):

    PROPERTY_NAME = "property_name"
    VALUE = "value"
    OLD_VALUE = "old_value"


class PropertyType(object):

    DEFAULT = 'p'
    EXTENDED = 'e'


class _JsonProperty(object):

    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    def json_value(self, *args, **kwargs):

        if inspect.isfunction(self.value):
            if len(inspect.signature(self.value).parameters) == 0:
                self._computed = self.value()
            else:
                self._computed = self.value(*args, **kwargs)
            return json.dumps(self._computed)
        return json.dumps(self.value)

    def get_value(self, *args, **kwargs):
        if hasattr(self, '_computed'):
            return self._computed
        if inspect.isfunction(self.value):
            if len(inspect.signature(self.value).parameters) == 0:
                return self.value()
            else:
                return self.value(*args, **kwargs)
        return self.value


class SubjectProperty(_JsonProperty):

    def __init__(self, name, value=None):
        super().__init__(name, value)
        self.type = PropertyType.DEFAULT


class SubjectExtProperty(_JsonProperty):

    def __init__(self, name, value=None):
        super().__init__(name, value)
        self.type = PropertyType.EXTENDED



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import rx
from dependency_injector import providers
from krules_core import RuleConst
from krules_core.base_functions import UpdatePayload, SetPayloadProperties, SetPayloadProperty, SetSubjectProperty, \
    SubjectPropertyChanged, SetSubjectExtendedProperty, SetSubjectPropertySilently, StoreSubjectProperty, \
    RuleFunctionBase, StoreSubjectPropertySilently, SetSubjectProperties, IncrementSubjectProperty, \
    DecrementSubjectProperty, IncrementSubjectPropertySilently, DecrementSubjectPropertySilently

from krules_core.core import RuleFactory
from .. import get_value_from_payload_diffs

from krules_core.providers import (
    event_router_factory,
    proc_events_rx_factory,
    subject_factory,
    subject_storage_factory)

counter = 0
asserted = []


@pytest.fixture
def subject():
    global counter
    counter += 1

    return subject_factory('test-subject-{0}'.format(counter)).flush()


@pytest.fixture
def router():
    router = event_router_factory()
    router.unregister_all()
    proc_events_rx_factory.override(providers.Singleton(rx.subjects.ReplaySubject))

    return event_router_factory()


@pytest.fixture
def asserted():
    global asserted
    asserted = []
    return asserted


filters = RuleConst.FILTERS
processing = RuleConst.PROCESSING
rulename = RuleConst.RULENAME
processed = RuleConst.PROCESSED


def _assert(name, expr, msg="test failed"):
    global asserted
    assert expr, msg
    asserted.append(name)


def test_payload_functions(subject, router, asserted):
    payload = {
        "k1": "val1",
        "k2": {"k2a": 1,
               "k2b": {"a": 1, "b": 2}}
    }

    RuleFactory.create(
        "test-alter-payload",
        subscribe_to="test-alter-payload",
        data={
            processing: [
                UpdatePayload({
                    "k2": {"k2b": {"b": 3, "c": 4}},
                    "k3": 3,
                }),
                SetPayloadProperties(k1=0, k3=lambda v: v+1, k4=lambda *args: len(args) and args[0]+1 or -1),
                SetPayloadProperty("k4", lambda *args: len(args) and args[0]+1 or -1)
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-alter-payload" and _assert(
            "test-update-1",
            get_value_from_payload_diffs("k3", x[processing][0]["payload_diffs"]) == 3 and
            "a" in x["payload"]["k2"]["k2b"] and x["payload"]["k2"]["k2b"]["a"] == payload["k2"]["k2b"]["a"] and
            not get_value_from_payload_diffs("k2/k2b/a", x[processing][0]["payload_diffs"]) and
            get_value_from_payload_diffs("k2/k2b/b", x[processing][0]["payload_diffs"]) == 3 and
            get_value_from_payload_diffs("k2/k2b/c", x[processing][0]["payload_diffs"]) == 4 and
            not get_value_from_payload_diffs("k2", x[processing][0]["payload_diffs"], default_value=None)
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-alter-payload" and _assert(
            "test-update-2",
            get_value_from_payload_diffs("k1", x[processing][1]["payload_diffs"]) == 0 and
            get_value_from_payload_diffs("k3", x[processing][1]["payload_diffs"]) == 4 and
            get_value_from_payload_diffs("k4", x[processing][1]["payload_diffs"]) == -1
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-alter-payload" and _assert(
            "test-update-3",
            get_value_from_payload_diffs("k4", x[processing][1]["payload_diffs"]) == 0
        )
    )

    router.route("test-alter-payload", subject, payload)

    assert "test-update-1" in asserted
    assert "test-update-2" in asserted


def test_subject_functions(subject, router, asserted):

    class _CheckStoredValue(RuleFunctionBase):

        def execute(self, prop, expected_value):
            if not subject_storage_factory(self.subject.name).is_concurrency_safe():
                return True   # skip test
            subject = subject_factory(self.subject.name)
            return subject.get(prop, cached=False) == expected_value


    from datetime import datetime
    RuleFactory.create(
        "test-set-subject-property",
        subscribe_to="test-set-subject-property",
        data={
            processing: [
                SetSubjectProperty("dt_prop", lambda: datetime.now().isoformat()),  # no args
                SetSubjectProperty("my_prop", 1),
                SetSubjectProperty("my_prop", lambda v: v+10),
                SetSubjectPropertySilently("something_to_say", False),
                StoreSubjectProperty("my_prop_2", 2),
                _CheckStoredValue("my_prop_2", 2),
                StoreSubjectPropertySilently("my_prop_3", 3),
                _CheckStoredValue("my_prop_3", 3),
                SetSubjectExtendedProperty("my_ext_prop", "extpropvalue"),
                SetSubjectProperties({
                    "my_prop_4": 4,
                    "my_silent_prop_5": 5
                }, unmuted=["my_prop_4"]),
                IncrementSubjectProperty("my_prop_4"),
                _CheckStoredValue("my_prop_4", 1),  # cached property in not considered
                DecrementSubjectProperty("my_prop_4", amount=1.5),
                IncrementSubjectPropertySilently("my_silent_prop_6"),
                DecrementSubjectPropertySilently("my_silent_prop_6"),
            ]
        }
    )
    from krules_core import types
    RuleFactory.create(
        "test-non-muted-property",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("my_prop", lambda value, old_value: value == 1 and old_value is None)
            ]
        }
    )
    RuleFactory.create(
        "test-muted-property",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("something_to_say")
            ]
        }
    )
    RuleFactory.create(
        "test-direct-property",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("my_prop_2", lambda value, old_value: value == 2 and old_value is None)
            ]
        }
    )
    RuleFactory.create(
        "test-muted-direct-property",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("my_prop_3")
            ]
        }
    )
    RuleFactory.create(
        "test-multi-set-properties-unmuted",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("my_prop_4")
            ]
        }
    )
    RuleFactory.create(
        "test-multi-set-properties-muted",   # never processed
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged(lambda p: p in ("my_silent_prop_5", "my_silent_prop_6"))
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-set-subject-property" and _assert(
            x[rulename],
            x[processed]
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-non-muted-property" and x[processed] and _assert(
            x[rulename],
            True
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-muted-property" and x[processed] and _assert(
            x[rulename],
            False
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-direct-property" and x[processed] and _assert(
            x[rulename],
            True
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-muted-direct-property" and x[processed] and _assert(
            x[rulename],
            False
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-multi-set-properties-unmuted" and x[processed] and _assert(
            x[rulename],
            True
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-multi-set-properties-muted" and x[processed] and _assert(
            x[rulename],
            False
        )
    )

    router.route("test-set-subject-property", subject, {})

    assert subject.get("my_prop") == 11
    assert subject.get("dt_prop")[:10] == datetime.now().isoformat()[:10]
    assert subject.get_ext("my_ext_prop") == "extpropvalue"
    assert subject.get("my_prop_4") == -.5
    assert subject.get("my_silent_prop_5") == 5

    assert "test-set-subject-property" in asserted
    assert "test-non-muted-property" in asserted
    assert "test-muted-property" not in asserted
    assert "test-direct-property" in asserted
    assert "test-multi-set-properties-unmuted" in asserted


# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import pytest
from dependency_injector import providers as providers


from krules_core.subject import PayloadConst, PropertyType

_test_events = []


class Router(object):

    def route(self, type, subject, payload):
        _test_events.append((type, subject, payload))


def setup_module(_):
    from dependency_injector import providers as providers
    from krules_core.providers import event_router_factory

    event_router_factory.override(
        providers.Singleton(Router)
    )


def teardown_module(_):
    from krules_core.providers import event_router_factory

    event_router_factory.reset_last_overriding()


counter = 0


# @pytest.fixture
# def messages():
#     from krules_core import ConfigKeyConst
#     from krules_core.providers import settings_factory
#     settings_factory.override(
#         providers.Singleton(
#             lambda: {
#                 ConfigKeyConst.MESSAGE_TOPICS_PREFIX: "tests-"
#             }
#         )
#     )
#     from krules_core import messages
#     return messages

def test_factory():

    from krules_core.providers import subject_factory
    test_subject = subject_factory("test-subject")
    assert test_subject.name == "test-subject"


@pytest.fixture
def subject():
    from krules_core.providers import subject_factory

    global counter
    counter += 1
    return subject_factory('test-subject-{0}'.format(counter)).flush()

@pytest.fixture
def subject_no_cache():
    from krules_core.providers import subject_factory

    global counter
    counter += 1
    return subject_factory('test-subject-{0}'.format(counter), use_cache_default=False).flush()


def test_set_get_del(subject):
    from krules_core.providers import subject_factory, subject_storage_factory

    subject.flush()

    # store nothing
    subject.store()

    global _test_events
    _test_events = []

    # USE CACHE (default)
    subject.set("my-prop", 1)
    assert subject.get("my-prop") == 1

    # not yet stored
    subject_copy = subject_factory(subject.name)
    with pytest.raises(AttributeError):
        subject_copy.get("my-prop")
    subject.store()
    subject_copy = subject_factory(subject.name)
    if subject_storage_factory(subject_copy.name).is_persistent():
        assert subject_copy.get("my-prop") == 1

    # callables
    new_value, old_value = subject.set("my-prop", lambda x: x+2)
    assert new_value == 3 and old_value == 1
    new_value, old_value = subject.set("my-prop-2", lambda x: x is None and 1 or x + 2)
    assert new_value == 1 and old_value is None

    # events
    assert len(_test_events) == 3
    from krules_core import types
    #   type
    assert _test_events[0][0] == _test_events[1][0] == _test_events[2][0] == types.SUBJECT_PROPERTY_CHANGED
    #   subject
    assert _test_events[0][1].name == _test_events[1][1].name == _test_events[2][1].name == subject.name
    #   payload
    payload = _test_events[0][2]
    assert payload[PayloadConst.PROPERTY_NAME] == "my-prop" and payload[PayloadConst.OLD_VALUE] is None \
        and payload[PayloadConst.VALUE] == 1
    payload = _test_events[1][2]
    assert payload[PayloadConst.PROPERTY_NAME] == "my-prop" and payload[PayloadConst.OLD_VALUE] == 1 \
        and payload[PayloadConst.VALUE] == 3
    payload = _test_events[2][2]
    assert payload[PayloadConst.PROPERTY_NAME] == "my-prop-2" and payload[PayloadConst.OLD_VALUE] is None \
        and payload[PayloadConst.VALUE] == 1

    # old the same with extended properties except that they are mute
    _test_events = []
    subject.set_ext("my-prop", 1)
    assert subject.get_ext("my-prop") == 1

    # not yet stored
    subject_copy = subject_factory(subject.name)
    with pytest.raises(AttributeError):
        subject_copy.get_ext("my-prop")
    subject.store()
    if subject_storage_factory(subject.name).is_persistent():
        subject_copy = subject_factory(subject.name)
        assert subject_copy.get_ext("my-prop") == 1

    # callables
    new_value, old_value = subject.set_ext("my-prop", lambda x: x+2)
    assert new_value == 3 and old_value == 1
    new_value, old_value = subject.set_ext("my-prop-2", lambda x: x is None and 1 or x + 2)
    assert new_value == 1 and old_value is None

    # events
    assert len(_test_events) == 0

    # muted
    val, _ = subject.set("my-prop", "silent", muted=True)
    assert val == "silent" and len(_test_events) == 0

    # NO CACHE
    subject.set("my-prop", 0, cached=False)  # not calling store
    if subject_storage_factory(subject.name).is_persistent():
        subject = subject_factory(subject.name)
        val = subject.get("my-prop", cached=False)
        assert val == 0
    #  with callables
    val, old_val = subject.set("my-prop", lambda x: x+5, cached=False)
    assert old_val == 0 and val == 5
    #  events produced
    assert len(_test_events) == 2

    # same with exts
    _test_events = []
    subject.set_ext("my-prop", 0, cached=False)  # not calling store
    if subject_storage_factory(subject.name).is_persistent():
        subject = subject_factory(subject.name)
        val = subject.get_ext("my-prop", cached=False)
        assert val == 0
    #  with callables
    val, old_val = subject.set_ext("my-prop", lambda x: x+5, cached=False)
    assert old_val == 0 and val == 5
    #  events NOT produced
    assert len(_test_events) == 0

    #  with cached
    if subject_storage_factory(subject.name).is_persistent():
        subject = subject_factory(subject.name)
        val = subject.get("my-prop")
        assert val == 5  # prop cached
        subject.set("my-prop", 1, cached=False)
        val = subject.get("my-prop")  # from cache
        assert val == 1
        subject.store()
        #  update cache
        subject = subject_factory(subject.name)
        val = subject.get("my-prop")
        assert val == 1  # prop cached
        subject.set("my-prop", 8, cached=True)
        val = subject.get("my-prop", cached=False)  # update cache
        assert val == 1
        val = subject.get("my-prop", cached=True)
        assert val == 1
        subject.store()

    # deletes
    _test_events = []
    #   cache not loaded yet
    subject.delete("my-prop", cached=False)
    assert len(_test_events) == 1 and \
        _test_events[0][0] == types.SUBJECT_PROPERTY_DELETED and \
        _test_events[0][1].name == subject.name and \
        _test_events[0][2][PayloadConst.PROPERTY_NAME] == "my-prop"
    with pytest.raises(AttributeError):
        subject.get("my-prop")
    with pytest.raises(AttributeError):
        subject.delete("my-prop")
    # add prop bypassing cache
    subject.set("my-prop", 0, cached=False)
    subject.delete("my-prop", cached=True)
    subject.store()
    with pytest.raises(AttributeError):
        subject.get("my-prop")
    # add prop in cache remove directly
    subject.set("my-prop", 0, cached=True)
    subject.delete("my-prop", cached=False)
    subject.store()
    with pytest.raises(AttributeError):
        subject.get("my-prop")
    # all in cache
    subject.set("my-prop", 0, cached=True)
    subject.delete("my-prop", cached=True)
    subject.store()
    with pytest.raises(AttributeError):
        subject.get("my-prop")
    # no cache
    subject.set("my-prop", 0, cached=False)
    subject.delete("my-prop", cached=False)
    subject.store()
    with pytest.raises(AttributeError):
        subject.get("my-prop")


def test_get_ext_props(subject):

    # no cache
    subject.set("my-prop-1", 0, cached=False)
    subject.set_ext("my-prop-2", 2, cached=False)
    subject.set_ext("my-prop-3", 3, cached=False)

    props = subject.get_ext_props()
    assert len(props) == 2
    assert "my-prop-2" in props and props["my-prop-2"] == 2
    assert "my-prop-3" in props and props["my-prop-3"] == 3

    # with cache
    subject.set_ext("my-prop-4", 4, cached=True)
    subject.set("my-prop-5", 5, cached=True)
    props = subject.get_ext_props()
    assert len(props) == 3
    assert "my-prop-2" in props and props["my-prop-2"] == 2
    assert "my-prop-3" in props and props["my-prop-3"] == 3
    assert "my-prop-4" in props and props["my-prop-4"] == 4


def test_cache_policy(subject):
    from krules_core.providers import subject_factory, subject_storage_factory

    # default policy: use cache
    subject.flush()
    subject.set("p1", None)
    subject.set_ext("p2", None)

    subject_fresh = subject_factory(subject.name)
    with pytest.raises(AttributeError):
        subject_fresh.get("p1")
    with pytest.raises(AttributeError):
        subject_fresh.get_ext("p2")

    subject = subject_factory(subject.name, use_cache_default=False)
    subject.flush()
    subject.set("p1", None)
    subject.set_ext("p2", None)

    if subject_storage_factory(subject.name).is_persistent():
        subject_fresh = subject_factory(subject.name)
        subject_fresh.get("p1")
        subject_fresh.get_ext("p2")


# def test_incr_decr(subject):
#     from krules_core import messages
#     from krules_core.providers import subject_factory, subject_storage_factory
#
#     global _test_events
#     _test_events = []
#     # no cached values
#     subject.flush()
#     # no initial value
#     subject.incr("my-prop")
#     # already stored
#     if subject_storage_factory(subject.name).is_persistent():
#         assert subject_factory(subject.name).get("my-prop") == 1
#     # cache updated
#     assert subject.get("my-prop") == 1
#     subject.incr("my-prop", 2)
#     assert subject.get("my-prop") == 3
#     if subject_storage_factory(subject.name).is_persistent():
#         assert subject_factory(subject.name).get("my-prop") == 3
#     # forks with float
#     subject.incr("my-prop", .1)
#     assert subject.get("my-prop") == 3.1
#     # decr
#     subject.decr("my-prop", .1)
#     assert subject.get("my-prop") == 3
#
#     # check events
#     assert len(_test_events) == 4
#     expected_values = (
#         (messages.SUBJECT_PROPERTY_CHANGED, subject, {'old_value': None, 'property_name': 'my-prop', 'value': 1}),
#         (messages.SUBJECT_PROPERTY_CHANGED, subject, {'old_value': 1, 'property_name': 'my-prop', 'value': 3}),
#         (messages.SUBJECT_PROPERTY_CHANGED, subject, {'old_value': 3, 'property_name': 'my-prop', 'value': 3.1}),
#         (messages.SUBJECT_PROPERTY_CHANGED, subject, {'old_value': 3.1, 'property_name': 'my-prop', 'value': 3.0}),
#     )
#     for ev in range(len(_test_events)):
#         for i in range(len(_test_events[ev])):
#             assert expected_values[ev][i] == _test_events[ev][i]
#
#     # check mute
#     subject.incr("my-prop", muted=True)
#     subject.decr("my-prop", muted=True)
#     assert len(_test_events) == 4


def test_len(subject):

    subject.flush()
    subject.set("my-prop-1", 1)
    subject.set("my-prop-2", 2)
    subject.set_ext("my-prop-3", 3)

    assert len(subject) == 2

    subject.flush()
    assert len(subject) == 2


def test_iter_contains(subject):

    subject.flush()
    subject.set("my-prop-1", 1)
    subject.set("my-prop-2", 2)
    subject.set_ext("my-prop-3", 3)

    count = 0
    for k in subject:
        assert k in ("my-prop-1", "my-prop-2")
        count += 1
    assert count == 2

    for p in ("my-prop-1", "my-prop-2"):
        assert p in subject


def test_property_proxy(subject_no_cache):

    from krules_core.providers import subject_factory, subject_storage_factory
    global _test_events
    _test_events = []

    subject_no_cache.flush()

    with pytest.raises(AttributeError):
        _ = subject_no_cache.foo

    subject_no_cache.foo = 1
    assert subject_no_cache.foo == 1
    assert len(_test_events) == 1

    if subject_storage_factory(subject_no_cache.name).is_persistent():
        assert subject_factory(subject_no_cache.name).get("foo") == 1

    assert subject_no_cache.m_foo == 1
    assert subject_no_cache.foo == 1
    subject_no_cache.m_foo = 2
    assert len(_test_events) == 1

    subject_no_cache.ext_foo = 3
    assert subject_no_cache.ext_foo == 3
    assert subject_no_cache.foo == 2
    assert len(_test_events) == 1
    if subject_storage_factory(subject_no_cache.name).is_persistent():
        assert subject_factory(subject_no_cache.name).get_ext("foo") == 3

    subject_no_cache.foo = lambda foo: foo * foo
    assert len(_test_events) == 2

    assert subject_no_cache.foo == 4

    subject_no_cache.foo.incr()
    assert len(_test_events) == 3
    assert subject_no_cache.foo == 5
    subject_no_cache.m_foo.incr(2)
    assert len(_test_events) == 3
    assert subject_no_cache.foo == 7
    subject_no_cache.foo.decr()
    assert subject_no_cache.foo == 6
    with pytest.raises(TypeError):
        subject_no_cache.ext_foo.incr()
    with pytest.raises(TypeError):
        subject_no_cache.ext_foo.decr()

    del subject_no_cache.foo
    assert len(_test_events) == 5
    with pytest.raises(AttributeError):
        _ = subject_no_cache.foo
    subject_no_cache.foo = 1
    del subject_no_cache.m_foo
    assert len(_test_events) == 6
    with pytest.raises(AttributeError):
        _ = subject_no_cache.foo
    del subject_no_cache.ext_foo
    with pytest.raises(AttributeError):
        _ = subject_no_cache.ext_foo


def test_reading(subject):

    stringvalue = "abcde"
    intvalue = 8
    dictvalue = {"a": 1, "b": 2}
    tuplevalue = (1, 2, 3, 4)
    listvalue = [1, 2, 3, 4]

    subject.stringvalue = stringvalue
    subject.intvalue = intvalue
    subject.dictvalue = dictvalue
    subject.tuplevalue = tuplevalue
    subject.listvalue = listvalue
    subject.store()
    subject._load()

    assert subject.stringvalue == stringvalue
    assert subject.intvalue == intvalue
    assert subject.dictvalue == dictvalue
    assert subject.tuplevalue == list(tuplevalue)
    assert subject.listvalue == listvalue





# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from krules_core.base_functions import RuleFunctionBase


class PyCall(RuleFunctionBase):
    """
    Execute a generic python function
    """

    def execute(self, func, args=(), kwargs={}, on_success=None, on_error=None,
                payload_dest="pycall_returns", raise_on_error=True):
        """
        Args:
            func: function to call
            args: function *args
            kwargs: function **kwargs
            on_success: function receiving (self, func_return_value)
            on_error: callable, if func raises an exception this function is called receiving (self, exception)
            payload_dest: key in payload storing func return value
            raise_on_error: if False and func produce an exception, it is not raised
        """

        try:
            ret = func(*args, **kwargs)
            if on_success is not None:
                on_success(self, ret)
            self.payload[payload_dest] = ret
        except Exception as exc:
            if on_error is not None:
                on_error(self, exc)
            if raise_on_error:
                raise exc





import rx

from dependency_injector import providers
from krules_core.providers import event_router_factory, proc_events_rx_factory

from krules_core.core import RuleFactory
from krules_core import RuleConst

from datetime import datetime
import logging


def _assert(expr, msg="test failed"):
    assert expr, msg
    return True



# TODO: write tests for dispatch policy

def test_router():
    subject = "test-subject"
    proc_events_rx_factory.override(providers.Singleton(rx.subjects.ReplaySubject))
    proc_events_rx = proc_events_rx_factory()

    start_time = datetime.now()
    router = event_router_factory()
    router.unregister_all()

    RuleFactory.create('test-empty-rule',
                       subscribe_to="some-type",
                       data={})

    proc_events_rx.subscribe(
        lambda x: _assert(
                    x[RuleConst.TYPE] == "some-type" and
                    "key1" in x[RuleConst.PAYLOAD] and
                    x[RuleConst.PAYLOAD]["key1"] == "val1",
        )
    )
    router.route('some-type', subject, {"key1": "val1"})

    end_time = datetime.now()
    logging.getLogger().debug("######### {}".format(end_time-start_time))

    assert router.unregister_all() == 1, "Expected 1 element"





from dependency_injector import providers as providers

from krules_core.providers import subject_storage_factory
from krules_core.tests.subject.sqlite_storage import SQLLiteSubjectStorage


def test_memorydatabase():

    subject_storage_factory.override(
        providers.Factory(lambda x, **kwargs: SQLLiteSubjectStorage(x, ":memory:"))
    )
    assert not subject_storage_factory("test-subject").is_persistent()
    assert not subject_storage_factory("test-subject").is_concurrency_safe()




# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from krules_core.subject import PayloadConst

from krules_core.base_functions import RuleFunctionBase
import inspect


class Returns(RuleFunctionBase):
    """
    Simply returns expression
    """

    def execute(self, expression):
        """
        Args:
            expression: We expect expression to be something callable that can be evaluated
        """
        return expression


class IsTrue(Returns):
    """
    True if True
    """

    def execute(self, expression):
        """
        Args:
            expression: We expect expression to be something callable that can be evaluated as a boolean
        """

        return bool(super().execute(expression))


class IsFalse(Returns):
    """
    True if False
    """

    def execute(self, expression):
        """
        Args:
            expression: We expect expression to be something callable that can be evaluated as a boolean
        """

        return not bool(super().execute(expression))


class SubjectNameMatch(RuleFunctionBase):
    """
    Checks if the subject's name matches the **regular expression**
    """

    def execute(self, regex, payload_dest="subject_match"):
        """
        Args:
            regex: Check expression
            payload_dest: Name of the key in the payload where the value of any groups contained in the expression is saved
        """
        import re
        match = re.search(regex, self.subject.name)
        if match is None:
            return False
        self.payload[payload_dest] = match.groupdict()
        return True


class SubjectNameDoesNotMatch(SubjectNameMatch):
    """
    Opposite of CheckSubjectNameMatch
    """

    def execute(self, regex, **kwargs):  # TODO: best inheritage support (**kwargs satisfy base class signature)

        return not super().execute(regex)


class CheckSubjectProperty(RuleFunctionBase):
    """
    Check the value of a property in the subject
    If the property does not exists returns False
    """

    def execute(self, property_name, property_value=lambda _none_: None, extended=False, cached=True):
        """
        Args:
            property_name: The name of the property
            property_value: Value to compare. If omitted, only the presence of the property is checked.
              If a callable is provided, this is invoked (optionally) with the property value
            See tests for examples
            extended: If True, check extended property
            cached: If False it checks the actual value on the storage backend bypassing the cached value
        """
        if property_name not in self.subject:
            return False
        _get = extended and self.subject.get_ext or self.subject.get
        if inspect.isfunction(property_value):
            sign = inspect.signature(property_value)
            if str(sign) == '(_none_)':
                return True
            n_args = len(sign.parameters)
            args = []
            if n_args > 0:
                args.append(_get(property_name, cached=cached))
            # if n_args > 1:
            #     args.append(self.payload)
            return property_value(*args)
        return _get(property_name, cached=cached) == property_value


class PayloadMatch(RuleFunctionBase):
    """
    It allows to process the payload with a jsonpath expression to check its content and possibly isolate part of it
    in a target variable
    """

    def execute(self, jp_expr, match_value=lambda _none_: None, payload_dest=None, single_match=False):
        """
        Args:
            jp_expr: Jsonpath expression
            payload_dest: If specified store the epression match result in that key in payload
            match_value: It can be both the expected value of the jsonpath expression processing or a boolean function
            that handles the expression result.
            single_match: if True produce a single value as result, a list of values otherwise

        >>> payload = {
        >>>             "user": "admin",
        >>>             "data": [{"id": 1, "value": 200}, {"id": 2, "value": 90}, {"id": 3, "value": 250}]}
        >>>         }
        >>> PayloadMatch("$.user", "admin")
        >>> False
        >>> PayloadMatch("$.user", "admin", single_match=True)
        >>> True
        >>> PayloadMatch("$.data[?@.value>100]")
        >>> True
        >>> PayloadMatch("$.data[?@.value>100]", [1, 3])
        >>> False
        >>> PayloadMatch("$.data[?@.value>100]", lambda x: len(x) == 2)
        >>> True
        """

        import jsonpath_rw_ext as jp

        matched = False
        fn = jp.match
        if single_match:
            fn = jp.match1

        match = fn(jp_expr, self.payload)
        if match is not None and len(match):
            matched = True

        if payload_dest:
            self.payload[payload_dest] = match

        if inspect.isfunction(match_value):
            sign = inspect.signature(match_value)
            if str(sign) != '(_none_)':
                n_args = len(sign.parameters)
                args = []
                if n_args > 0:
                    args.append(match)

                matched = match_value(*args)
        else:
            matched = match == match_value

        return matched


class PayloadMatchOne(PayloadMatch):
    """
    Same as CheckPayloadJPMatch but expects just one element as result
    """

    def execute(self, jp_expr, match_value=lambda _none_: None, payload_dest=None, **kwargs):
        """
        Args:
            jp_expr: Jsonpath expression
            payload_dest: Destination key in payload
        """
        return super().execute(jp_expr, match_value, payload_dest, True)


class SubjectPropertyChanged(RuleFunctionBase):
    """
    Catch the event subject property changed
    """

    def execute(self, property_name, value=lambda _none_: None, old_value=lambda _none_: None):
        """
        Args:
            property_name: Name of property changed. Accept callable receiving (optionally) the property name, In that
               case it must returns a boolean
            value: If not a callable the value is compared, otherwise it can receive no parameters,
               or the value of the property and eventually also the previous value of the property (old_value).
               It must returns a boolean
            old_value: If not a callable the value is compared, otherwise it receives (optionally) the old_value and
               must returns a boolean
        """

        # property_name
        if inspect.isfunction(property_name):
            sign = inspect.signature(property_name)
            n_args = len(sign.parameters)
            if n_args == 0:
                matched = self.payload[PayloadConst.PROPERTY_NAME] == property_name()
            elif n_args == 1:
                matched = property_name(self.payload[PayloadConst.PROPERTY_NAME])
            else:
                raise TypeError("takes at most two arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.PROPERTY_NAME] == property_name

        if not matched:
            return False

        # value
        if inspect.isfunction(value):
            sign = inspect.signature(value)
            if str(sign) != '(_none_)':
                n_args = len(sign.parameters)
                if n_args == 0:
                    matched = self.payload[PayloadConst.VALUE] == value()
                elif n_args == 1:
                    matched = value(self.payload[PayloadConst.VALUE])
                elif n_args == 2:
                    args = [self.payload[PayloadConst.VALUE], self.payload[PayloadConst.OLD_VALUE]]  # for IDE happiness
                    matched = value(*args)
                else:
                    raise TypeError("takes at most three arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.VALUE] == value

        if not matched:
            return False

        # old_value
        if inspect.isfunction(old_value):
            sign = inspect.signature(old_value)
            if str(sign) != '(_none_)':
                n_args = len(sign.parameters)
                if n_args == 0:
                    matched = self.payload[PayloadConst.OLD_VALUE] == old_value()
                elif n_args == 1:
                    matched = old_value(self.payload[PayloadConst.OLD_VALUE])
                else:
                    raise TypeError("takes at most two arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.OLD_VALUE] == old_value

        if not matched:
            return False

        return True



class EmptySubjectStorage:

    def is_concurrency_safe(self):

        return False

    def is_persistent(self):

        return False

    def load(self):
        return {}, {}

    def store(self, inserts=[], updates=[], deletes=[]):
        pass

    def set(self, prop, old_value_default=None):

        return None, None

    def get(self, prop):

        return None

    def delete(self, prop):
        pass

    def get_ext_props(self):

        return {}

    def flush(self):

        return self



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import re

import pytest
import rx
from dependency_injector import providers
from krules_core.base_functions.filters import Returns, CheckSubjectProperty, PayloadMatch, SubjectNameMatch, \
    SubjectNameDoesNotMatch, IsTrue, IsFalse, PayloadMatchOne, SubjectPropertyChanged

from krules_core import RuleConst

from krules_core.core import RuleFactory

from krules_core.providers import (
    event_router_factory,
    proc_events_rx_factory,
    subject_factory,
)

counter = 0
asserted = []

@pytest.fixture
def subject():
    global counter
    counter += 1

    return subject_factory('test-subject-{0}'.format(counter)).flush()


@pytest.fixture
def router():
    router = event_router_factory()
    router.unregister_all()
    proc_events_rx_factory.override(providers.Singleton(rx.subjects.ReplaySubject))

    return event_router_factory()

@pytest.fixture
def asserted():
    global asserted
    asserted = []
    return asserted


filters = RuleConst.FILTERS
processing = RuleConst.PROCESSING
rulename = RuleConst.RULENAME
processed = RuleConst.PROCESSED


def _assert(name, expr, msg="test failed"):
    global asserted
    assert expr, msg
    asserted.append(name)


def test_return(subject, router, asserted):

    RuleFactory.create('test-returns',
                       subscribe_to="event-test-returns",
                       data={
                           filters: [
                               Returns("something"),
                           ],
                       })

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == 'test-returns' and _assert(
            'test-returns-something',
            x[processed] and x[filters][0]["returns"] == "something"
            )

    )

    router.route("event-test-returns", subject, {})

    assert 'test-returns-something' in asserted


def test_truth(subject, router, asserted):

    RuleFactory.create('test-is-true',
                       subscribe_to="event-test-truth",
                       data={
                           filters: [
                               IsTrue(
                                    lambda payload: payload["it-works"]
                               ),
                           ],
                       })

    RuleFactory.create('test-is-false',
                       subscribe_to="event-test-truth",
                       data={
                           filters: [
                               IsFalse(
                                   lambda payload: payload["it-works"] is False
                               ),
                           ],
                       })

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == 'test-is-true' and _assert(
            'test-is-true',
            x[processed] and x[filters][0]["returns"] is True
            )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == 'test-is-false' and _assert(
            'test-is-false',
            x[processed] and x[filters][0]["returns"] is True
            )

    )

    router.route("event-test-truth", subject, {'it-works': True})

    assert 'test-is-true' in asserted
    assert 'test-is-false' in asserted


def test_subject_match(router, asserted):

    user_subject = "user|000001"

    RuleFactory.create('test-subject-match',
                       subscribe_to='event-user-action',
                       data={
                           filters: [
                               SubjectNameMatch(r"^user\|(?P<user_id>.+)", payload_dest="user_info"),
                               IsTrue(
                                   lambda payload: "user_id" in payload.get("user_info", {})
                               )
                           ]
                       })

    RuleFactory.create('test-subject-does-not-match',
                       subscribe_to='event-user-action',
                       data={
                           filters: [
                               SubjectNameDoesNotMatch(r"^device\|(?P<device_id>.+)", payload_dest="device_info"),
                           ]
                       })

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == 'test-subject-match' and _assert(
            'test-subject-match',
            x[processed] is True
            )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == 'test-subject-does-not-match' and _assert(
            'test-subject-does-not-match',
            x[processed] is True
            )
    )

    router.route("event-user-action", user_subject, {})

    assert 'test-subject-match' in asserted
    assert 'test-subject-does-not-match' in asserted


def test_check_subject_property(router, subject, asserted):

    RuleFactory.create(
        "test-simple-subject-property",
        subscribe_to="test-subject-property",
        data={
            filters: [
                CheckSubjectProperty("prop-1", "value-1"),
                CheckSubjectProperty("prop-2", 2),
            ]
        }
    )

    RuleFactory.create(
        "test-simple-subject-property-fails",
        subscribe_to="test-subject-property",
        data={
            filters: [
                CheckSubjectProperty("prop-1", "value-1"),
                CheckSubjectProperty("prop-2", "2"),
            ]
        }
    )

    import re
    RuleFactory.create(
        "test-expr-subject-property",
        subscribe_to="test-subject-property",
        data={
            filters: [
                # one argument (value)
                CheckSubjectProperty("prop-1"),
                CheckSubjectProperty("prop-1", lambda v: v in ("value-1",) and re.match("value-[0-9]", v)),
                CheckSubjectProperty("prop-2", lambda v: type(v) is int),
                CheckSubjectProperty("ext-prop", extended=True)
            ]
        }
    )

    subject.set("prop-1", "value-1")
    subject.set("prop-2", 2)
    subject.set_ext("ext-prop-2", "extprop")

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-simple-subject-property" and _assert(
            x[rulename],
            x[processed] is True
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-simple-subject-property-fails" and _assert(
            x[rulename],
            x[processed] is False
        )
    )

    router.route("test-subject-property", subject, {})

    assert "test-simple-subject-property" in asserted
    assert "test-simple-subject-property-fails" in asserted

    # clean up
    router.unregister_all()
    proc_events_rx_factory.override(providers.Singleton(rx.subjects.ReplaySubject))

    # check direct
    subject.set("prop-1", "value-2")
    subject.set_ext("ext-prop-3", 1)

    # properties are not yet stored, rules should fail
    RuleFactory.create(
        "test-subject-property-direct",
        subscribe_to="test-subject-property-direct",
        data={
            filters: [
                CheckSubjectProperty("v-prop-1", "value-2", cached=False),  # prop-1 is still value-1
            ]
        }
    )

    RuleFactory.create(
        "test-subject-property-ext-direct",
        subscribe_to="test-subject-property-direct",
        data={
            filters: [
                CheckSubjectProperty("ext-prop-3", cached=False),  # ext-prop-3 does not exists yet
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-subject-property-direct" and _assert(
            x[rulename],
            not x[processed]
        )
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-subject-property-ext-direct" and _assert(
            x[rulename],
            not x[processed]
        )
    )

    router.route("test-subject-property-direct", subject, {})

    assert "test-subject-property-direct" in asserted
    assert "test-subject-property-ext-direct" in asserted


def test_check_payload_match(router, subject, asserted):

    payload = {
        "batch_data": [
          {
            "time": "2020-03-10T17:16:16.014673",
            "value": 100
          },
          {
            "time": "2020-03-10T17:21:16.014673",
            "value": 50
          },
          {
            "time": "2020-03-10T17:26:16.014673",
            "value": 60
          },
          {
            "time": "2020-03-10T17:31:16.014673",
            "value": 105
          },
          {
            "time": "2020-03-10T17:36:16.014673",
            "value": 120
          }
        ],
        "device_info": {
            "id": "0AFB1110",
            "disabled": False,
        }
      }

    # just check no empty
    RuleFactory.create(
        "test-check-payload-jpmatch-not-empty",
        subscribe_to="test-check-payload-jpmatch",
        data={
            filters: [
                PayloadMatch("$..batch_data[?@.value>100]")  # returns two elements - pass
            ]
        }
    )
    # store result
    RuleFactory.create(
        "test-check-payload-jpmatch-store-result",
        subscribe_to="test-check-payload-jpmatch",
        data={
            filters: [
                PayloadMatch("$.batch_data[?@.value>100]", lambda m: len(m) == 2),
                PayloadMatch("$.batch_data[?@.value>100]", payload_dest="jpexpr_match"),
                IsTrue(
                    lambda payload: len(payload['jpexpr_match']) == 2
                )
            ]
        }
    )
    # match one
    RuleFactory.create(
        "test-check-payload-jpmatch-one",
        subscribe_to="test-check-payload-jpmatch",
        data={
            filters: [
                PayloadMatchOne("$.device_info.id", "0AFB1110"),
                PayloadMatchOne("$.device_info.id", payload_dest="device_id"),
                PayloadMatchOne("$.device.info.disabled", lambda disabled: not disabled),
                IsTrue(
                    lambda payload: payload["device_id"] == "0AFB1110"
                )
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-check-payload-jpmatch-not-empty" and _assert(
            x[rulename],
            x[processed]
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-check-payload-jpmatch-store-result" and _assert(
            x[rulename],
            x[processed]
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-check-payload-jpmatch-one" and _assert(
            x[rulename],
            x[processed]
            and x[filters][0]["returns"] is True
            and x[filters][1]["returns"] is True
            and x[filters][2]["returns"] is True
        )
    )

    router.route("test-check-payload-jpmatch", subject, payload)

    assert "test-check-payload-jpmatch-not-empty" in asserted
    assert "test-check-payload-jpmatch-store-result" in asserted
    assert "test-check-payload-jpmatch-one" in asserted


def test_on_subject_property_changed(router, subject, asserted):

    from krules_core import types
    RuleFactory.create(
        "test-prop-changed",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("prop_a"),
                SubjectPropertyChanged(lambda: "prop_{}".format("a")),
                SubjectPropertyChanged(lambda prop: re.match("prop_[a-z]", prop)),
                SubjectPropertyChanged("prop_a", lambda: 1),
                SubjectPropertyChanged("prop_a", value=lambda value: value > 0),
                SubjectPropertyChanged("prop_a", value=lambda value, old_value: value == 1 and old_value is None),
                SubjectPropertyChanged("prop_a", old_value=None),
                SubjectPropertyChanged("prop_a", old_value=lambda old_value: old_value is None)
           ]
        }
    )
    RuleFactory.create(
        "test-prop-changed-fails-1",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("prop_b"),
           ]
        }
    )
    RuleFactory.create(
        "test-prop-changed-fails-2",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("prop_a", lambda value: value > 1),
            ]
        }
    )
    RuleFactory.create(
        "test-prop-changed-fails-3",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("prop_a", lambda value, old_value: value == 1 and old_value is not None),
            ]
        }
    )
    RuleFactory.create(
        "test-prop-changed-fails-4",
        subscribe_to=types.SUBJECT_PROPERTY_CHANGED,
        data={
            filters: [
                SubjectPropertyChanged("prop_a", lambda value, old_value: value == 1,
                                         lambda old_value: old_value is not None),
            ]
        }
    )

    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-prop-changed" and _assert(
            x[rulename],
            x[processed], "{} not processed".format(x[rulename])
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-prop-changed-fails-1" and _assert(
            x[rulename],
            not x[processed], "{} should not be not processed".format(x[rulename])
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-prop-changed-fails-2" and _assert(
            x[rulename],
            not x[processed], "{} should not be not processed".format(x[rulename])
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-prop-changed-fails-3" and _assert(
            x[rulename],
            not x[processed], "{} should not be not processed".format(x[rulename])
        )
    )
    proc_events_rx_factory().subscribe(
        lambda x: x[rulename] == "test-prop-changed-fails-4" and _assert(
            x[rulename],
            not x[processed], "{} should not be not processed".format(x[rulename])
        )
    )


    subject.prop_a = 1

    assert 'test-prop-changed' in asserted
    assert 'test-prop-changed-fails-1' in asserted
    assert 'test-prop-changed-fails-2' in asserted
    assert 'test-prop-changed-fails-3' in asserted
    assert 'test-prop-changed-fails-4' in asserted






# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .providers import configs_factory
from . import ConfigKeyConst


_settings = configs_factory()


def format_event_type(type):
    return "{}{}".format(_settings.get(ConfigKeyConst.TYPE_TOPICS_PREFIX, ""), type)


_m = format_event_type


SUBJECT_PROPERTY_CHANGED = _m("subject-property-changed")
SUBJECT_PROPERTY_DELETED = _m("subject-property-deleted")



import inspect

import wrapt

from krules_core.subject import SubjectProperty, SubjectExtProperty, PayloadConst, PropertyType


class Subject(object):

    """
    Subject implementation
    Needs a storage strategy implementation
    """

    def __init__(self, name, event_info={}, event_data=None, use_cache_default=True):
        from krules_core.providers import subject_storage_factory

        self.name = name
        self._use_cache = use_cache_default
        self._storage = subject_storage_factory(name, event_info=event_info, event_data=event_data)
        self._event_info = event_info
        self._cached = None

    def __str__(self):

        return self.name

    def _load(self):

        props, ext_props = self._storage.load()
        #if self._cached is None:
        self._cached = \
            {
                PropertyType.DEFAULT: {
                    "values": {},
                    "created": set(),
                    "updated": set(),
                    "deleted": set(),
                },
                PropertyType.EXTENDED: {
                    "values": {},
                    "created": set(),
                    "updated": set(),
                    "deleted": set(),
                }
            }
        self._cached[PropertyType.DEFAULT]["values"] = props
        self._cached[PropertyType.EXTENDED]["values"] = ext_props

    def _set(self, prop, value, extended, muted, cached):
        if isinstance(value, tuple):
            value = list(value)

        if cached is None:
            cached = self._use_cache
        if cached:
            if self._cached is None:
                self._load()
            kprops = extended and PropertyType.EXTENDED or PropertyType.DEFAULT
            vals = extended and self._cached[kprops]["values"] or self._cached[kprops]["values"]
            if prop in vals:
                self._cached[kprops]["updated"].add(prop)
            else:
                self._cached[kprops]["created"].add(prop)
            try:
                old_value = vals[prop]
            except KeyError:
                old_value = None
            if inspect.isfunction(value):
                n_params = len(inspect.signature(value).parameters)
                if n_params == 0:
                    value = value()
                elif n_params == 1:
                    value = value(old_value)
                else:
                    raise ValueError("to many arguments for {}".format(prop))

            vals[prop] = value
        else:
            klass, k = extended and (SubjectExtProperty, PropertyType.EXTENDED) or (SubjectProperty, PropertyType.DEFAULT)
            value, old_value = self._storage.set(klass(prop, value))
            # update cached
            if self._cached:
                self._cached[k]["values"][prop] = value
                if prop in self._cached[k]["created"]:
                    self._cached[k]["created"].remove(prop)
                if prop in self._cached[k]["updated"]:
                    self._cached[k]["updated"].remove(prop)
                if prop in self._cached[k]["deleted"]:
                    self._cached[k]["deleted"].remove(prop)

        if not muted and value != old_value:
            payload = {PayloadConst.PROPERTY_NAME: prop, PayloadConst.OLD_VALUE: old_value,
                       PayloadConst.VALUE: value}

            from krules_core.providers import event_router_factory
            from krules_core import types
            event_router_factory().route(types.SUBJECT_PROPERTY_CHANGED, self, payload)

        return value, old_value

    def set(self, prop, value, muted=False, cached=None):
        return self._set(prop, value, False, muted, cached)

    def set_ext(self, prop, value, cached=None):
        return self._set(prop, value, True, True, cached)

    def _get(self, prop, extended, cached):
        if cached is None:
            cached = self._use_cache
        if cached:
            if self._cached is None:
                self._load()
            if extended:
                vals = self._cached[PropertyType.EXTENDED]["values"]
            else:
                vals = self._cached[PropertyType.DEFAULT]["values"]
            if prop not in vals:
                raise AttributeError(prop)
            return vals[prop]
        else:
            klass, k = extended and (SubjectExtProperty, PropertyType.EXTENDED) or (SubjectProperty, PropertyType.DEFAULT)
            val = self._storage.get(klass(prop))
            # update cache if present
            if self._cached is not None:
                self._cached[k]["values"][prop] = val
                # remove prop from inserts and ensure it is in updates (ignore deletes)
                if prop in self._cached[k]["created"]:
                    self._cached[k]["created"].remove(prop)
                self._cached[k]["updated"].add(prop)
            return val

    def get(self, prop, cached=None):
        return self._get(prop, False, cached)

    def get_ext(self, prop, cached=None):
        return self._get(prop, True, cached)

    def _delete(self, prop, extended, muted, cached):
        if cached is None:
            cached = self._use_cache
        if cached:
            if self._cached is None:
                self._load()
            k = extended and PropertyType.EXTENDED or PropertyType.DEFAULT
            vals = self._cached[k]["values"]
            if prop not in vals:
                raise AttributeError(prop)
            del vals[prop]
            for _set in ("created", "updated"):
                if prop in self._cached[k][_set]:
                    self._cached[k][_set].remove(prop)
            self._cached[k]["deleted"].add(prop)
        else:
            klass, k = extended and (SubjectExtProperty, PropertyType.EXTENDED) or (SubjectProperty, PropertyType.DEFAULT)
            self._storage.delete(klass(prop))
            if self._cached is not None:
                if prop in self._cached[k]["values"]:
                    del self._cached[k]["values"][prop]
                for _set in ["created", "updated", "deleted"]:
                    if prop in self._cached[k][_set]:
                        self._cached[k][_set].remove(prop)

        if not muted:
            payload = {PayloadConst.PROPERTY_NAME: prop}

            from krules_core.providers import event_router_factory
            from krules_core import types
            event_router_factory().route(types.SUBJECT_PROPERTY_DELETED, self, payload)

    def delete(self, prop, muted=False, cached=None):
        self._delete(prop, False, muted, cached)

    def delete_ext(self, prop, cached=None):
        self._delete(prop, True, False, cached)


    def get_ext_props(self):
        # If we have a cache we use it, otherwise we don't load any cache
        # and we get them from the storage.
        # This is because we need all the extended properties primarily when we route events to a subject
        # and we don't care about normal properties
        if self._cached:
            return self._cached[PropertyType.EXTENDED]["values"].copy()
        return self._storage.get_ext_props()

    def event_info(self):
        return self._event_info.copy()

    def flush(self):
        self._storage.flush()
        return self

    def store(self):

        if not self._cached:
            return

        inserts, updates, deletes = [], [], []
        for _set, k1 in ((inserts, "created"), (updates, "updated"), (deletes, "deleted")):
            for k2, klass in ((PropertyType.DEFAULT, SubjectProperty), (PropertyType.EXTENDED, SubjectExtProperty)):
                for prop in self._cached[k2][k1]:
                    try:
                        _set.append(klass(prop, self._cached[k2]["values"][prop]))
                    except KeyError as ex:
                        if _set is deletes:
                            _set.append(klass(prop))
                        else:
                            raise ex

        self._storage.store(inserts=inserts, updates=updates, deletes=deletes)
        self._cached = None

    def __len__(self):

        if self._cached is None or not self._use_cache:
            self._load()
        return len(self._cached[PropertyType.DEFAULT]["values"])

    def __iter__(self):
        if self._cached is None or not self._use_cache:
            self._load()
        return iter(self._cached[PropertyType.DEFAULT]["values"])

    def __contains__(self, item):
        if self._cached is None or not self._use_cache:
            self._load()
        return item in self._cached[PropertyType.DEFAULT]["values"]

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError as ex:
            propname = item
            is_ext = False
            is_mute = False
            try:
                if item.startswith("m_"):
                    propname = item[2:]
                    is_mute = True
                elif item.startswith("ext_"):
                    propname = item[4:]
                    is_ext = True

                value = self._get(propname, extended=is_ext, cached=self._use_cache)
            except KeyError:
                raise ex
            return _SubjectPropertyProxy(self, propname, value, is_ext, is_mute, self._use_cache)

    def __setattr__(self, item, value):

        if item in ('name',) or item.startswith("_"):
            return super().__setattr__(item, value)

        is_mute = False
        propname = item
        is_ext = False
        if item.startswith("m_"):
            is_mute = True
            propname = item[2:]
        elif item.startswith("ext_"):
            is_mute = True
            is_ext = True
            propname = item[4:]
        return self._set(propname, value, is_ext, is_mute, self._use_cache)

    def __delattr__(self, item):
        if item in ('name',) or item.startswith("_"):
            raise Exception("cannot remove {}".format(item))

        is_mute = False
        propname = item
        is_ext = False
        if item.startswith("m_"):
            is_mute = True
            propname = item[2:]
        elif item.startswith("ext_"):
            is_mute = True
            is_ext = True
            propname = item[4:]
        return self._delete(propname, is_ext, is_mute, self._use_cache)


class _SubjectPropertyProxy(wrapt.ObjectProxy):
        """
        This class wraps subject properties and it is ment primarily
        to use in interactive mode.
        I also provides convenience methods to dial with counters (incr/decr)
        All operations are not cached and immediately effective
        """

        _subject = None
        _prop = None
        _extended = None
        _muted = None
        _use_cache = None

        def __init__(self, subject, prop, value, extended, muted, use_cache):
            super().__init__(value)
            self._subject = subject
            self._prop = prop
            self._extended = extended
            self._muted = muted
            self._use_cache = use_cache

        def __repr__(self):
            return self.__class__.__repr__(self.__wrapped__)

        def incr(self, amount=1):
            if self._extended:
                raise TypeError("not supported for extended properties")
            return self._subject.set(self._prop, lambda v: v+amount, self._muted, self._use_cache)

        def decr(self, amount=1):
            if self._extended:
                raise TypeError("not supported for extended properties")
            return self._subject.set(self._prop, lambda v: v-amount, self._muted, self._use_cache)







# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from uuid import uuid4

from . import RuleConst as Const
from .providers import event_router_factory, subject_factory

import sys
import traceback


import logging
logger = logging.getLogger("__core__")

from .providers import exceptions_dumpers_factory
import jsonpatch

from collections.abc import Mapping

class Rule:

    def __init__(self, name, description=""):
        self.name = name
        self.description = description,
        self._filters = []
        self._processing = []
        self._finally = []

    def set_filters(self, filters):

        assert(isinstance(filters, type([])))
        self._filters.extend(filters)

    def set_processing(self, processing):
        assert(isinstance(processing, type([])))
        self._processing.extend(processing)

    def set_finally(self, finally_):
        self._finally.extend(finally_)

    def _process(self, event_type, subject, payload):

        def __get_signature_info(func):
            signature = inspect.signature(func)
            return "%s(%s)" % (func.__name__, ", ".join(signature.parameters))

        def __clean(dd):
            del dd[Const.PROCESS_ID]
            del dd[Const.TYPE]
            del dd[Const.SUBJECT]
            del dd[Const.RULENAME]
            del dd[Const.SECTION]
            dd.get(Const.PAYLOAD, {}).pop("_event_info", None)
            return dd

        def __copy_list(ll):
            dst = []
            for el in ll:
                if isinstance(el, Mapping):
                    dst.append(__copy(el))
                elif isinstance(el, (list, tuple)):
                    dst.append(__copy_list(el))
                elif inspect.isfunction(el):
                    dst.append(__get_signature_info(el))
                elif isinstance(el, (bool, int, float, str)) or el is None:
                    dst.append(el)
                else:
                    dst.append(str(el))
            return dst

        def __copy(pp):
            cp = {}
            for k, v in pp.items():
                if isinstance(v, Mapping):
                    cp[k] = {}
                    cp[k] = __copy(v)
                elif isinstance(v, (list, tuple)):
                    cp[k] = __copy_list(v)
                elif inspect.isfunction(v):
                    cp[k] = __get_signature_info(v)
                elif isinstance(v, (bool, int, float, str)) or v is None:
                    cp[k] = v
                else:
                    cp[k] = str(v)
            return cp

        logger.debug("process {0} for {1}".format(event_type, self.name))

        if isinstance(subject, str):
            subject = subject_factory(subject)

        from .providers import proc_events_rx_factory

        proc_events_rx = proc_events_rx_factory()  # one event for each processed rule

        process_id = str(uuid4())
        payload_copy = __copy(payload)
        event_info = payload_copy.pop("_event_info", {})

        res_full = {
            Const.TYPE: event_type,
            Const.SUBJECT: str(subject.name),
            Const.RULENAME: self.name,
            Const.PAYLOAD: payload_copy,
            Const.FILTERS: [],
            Const.PROCESSING: [],
            Const.GOT_ERRORS: False,
            Const.EVENT_INFO: event_info,
        }

        last_payload = __copy(payload)

        res_in = {}
        processed_args = {}
        processed_kwargs = {}
        try:
            for _c in self._filters:
                if inspect.isclass(_c):
                    _c = _c()
                _cinst_name = _c.__class__.__name__
                _cinst = type(_cinst_name, (_c.__class__,), {})()
                _cinst.type = event_type
                _cinst.subject = subject
                _cinst.payload = payload
                res_in = {
                    Const.PROCESS_ID: process_id,
                    Const.TYPE: event_type,
                    Const.SUBJECT: str(subject.name),
                    Const.RULENAME: self.name,
                    Const.SECTION: Const.FILTERS,
                    Const.FUNC_NAME: _cinst_name,
                    Const.PAYLOAD: __copy(payload),
                    Const.ARGS: __copy_list(_c._args),
                    Const.KWARGS: __copy(_c._kwargs),
                }
                logger.debug("> processing: {0}".format(res_in))
                try:
                    processed_args = _c._get_args(_cinst)
                    processed_kwargs = _c._get_kwargs(_cinst)
                    res = _cinst.execute(*processed_args, **processed_kwargs)
                except TypeError as ex:
                    msg = "{} in {}: ".format(_cinst_name, self.name)
                    raise TypeError(msg + str(ex))

                payload_copy = __copy(payload)
                payload_patches = jsonpatch.JsonPatch.from_diff(last_payload, payload_copy).patch
                last_payload = payload_copy
                res_out = {
                    Const.PROCESS_ID: res_in[Const.PROCESS_ID],
                    Const.TYPE: res_in[Const.TYPE],
                    Const.SUBJECT: res_in[Const.SUBJECT],
                    Const.RULENAME: res_in[Const.RULENAME],
                    Const.SECTION: res_in[Const.SECTION],
                    Const.FUNC_NAME: res_in[Const.FUNC_NAME],
                    Const.PAYLOAD_DIFFS: payload_patches,
                    Const.ARGS: __copy_list(processed_args),
                    Const.KWARGS: __copy(processed_kwargs),
                    Const.RETURNS: res
                }
                logger.debug("< processed: {0}".format({'payload_diffs': res_out[Const.PAYLOAD_DIFFS], 'returns': res_out[Const.RETURNS]}))
                res_full[Const.FILTERS].append(__clean(res_out))
                if not res:
                    res_full[Const.PROCESSED] = False
                    proc_events_rx.on_next(res_full)
                    return

            res_full[Const.PROCESSED] = True

            for _c in self._processing:
                if inspect.isclass(_c):
                    _c = _c()
                _cinst_name = _c.__class__.__name__
                _cinst = type(_cinst_name, (_c.__class__,), {})()
                _cinst.type = event_type
                _cinst.subject = subject
                _cinst.payload = payload
                res_in = {
                    Const.PROCESS_ID: process_id,
                    Const.TYPE: event_type,
                    Const.SUBJECT: str(subject.name),
                    Const.RULENAME: self.name,
                    Const.SECTION: Const.PROCESSING,
                    Const.FUNC_NAME: _cinst_name,
                    Const.PAYLOAD: __copy(payload),
                    Const.ARGS: __copy_list(_c._args),
                    Const.KWARGS: __copy(_c._kwargs),
                }
                logger.debug("> processing: {0}".format(res_in))
                try:
                    processed_args = _c._get_args(_cinst)
                    processed_kwargs = _c._get_kwargs(_cinst)
                    res = _cinst.execute(*processed_args, **processed_kwargs)
                except TypeError as ex:
                    msg = "{} in {}: ".format(_cinst_name, self.name)
                    raise TypeError(msg + str(ex))
                payload_copy = __copy(payload)
                payload_patches = jsonpatch.JsonPatch.from_diff(last_payload, payload_copy).patch
                last_payload = payload_copy
                res_out = {
                    Const.PROCESS_ID: res_in[Const.PROCESS_ID],
                    Const.TYPE: res_in[Const.TYPE],
                    Const.SUBJECT: res_in[Const.SUBJECT],
                    Const.RULENAME: res_in[Const.RULENAME],
                    Const.SECTION: res_in[Const.SECTION],
                    Const.FUNC_NAME: res_in[Const.FUNC_NAME],
                    Const.PAYLOAD_DIFFS: payload_patches,
                    Const.ARGS: __copy_list(processed_args),
                    Const.KWARGS: __copy(processed_kwargs),
                    Const.RETURNS: res,
                }
                logger.debug("< processed: {0}".format({'payload_diffs': res_out[Const.PAYLOAD_DIFFS], 'returns': res_out[Const.RETURNS]}))
                res_full[Const.PROCESSING].append(__clean(res_out))

            proc_events_rx.on_next(res_full)

        except Exception as e:
            logger.error("catched exception of type {0} ({1})".format(type(e), getattr(e, 'message', str(e))))
            if proc_events_rx:
                payload_copy = __copy(payload)
                payload_patches = jsonpatch.JsonPatch.from_diff(last_payload, payload_copy).patch
                type_, value_, traceback_ = sys.exc_info()
                res_out = {
                    Const.PROCESS_ID: res_in[Const.PROCESS_ID],
                    Const.TYPE: res_in[Const.TYPE],
                    Const.SUBJECT: res_in[Const.SUBJECT],
                    Const.RULENAME: res_in[Const.RULENAME],
                    Const.SECTION: res_in[Const.SECTION],
                    Const.FUNC_NAME: res_in[Const.FUNC_NAME],
                    Const.PAYLOAD_DIFFS: payload_patches,
                    Const.ARGS: __copy_list(processed_args),
                    Const.KWARGS: __copy(processed_kwargs),
                    Const.RETURNS: None,
                    Const.EXCEPTION: ".".join([type(e).__module__, type(e).__name__]),
                    Const.EXC_INFO: traceback.format_exception(type_, value_, traceback_),
                    Const.EXC_EXTRA_INFO: exceptions_dumpers_factory().dump(e),
                }
                logger.error(res_out)
                logger.debug("# unprocessed: {0}".format(res_out))
                res_full[Const.GOT_ERRORS] = True
                res_full[res_out[Const.SECTION]].append(__clean(res_out))
                proc_events_rx.on_next(res_full)


class RuleFactory:

    @staticmethod
    def create(name: object, description: object = "", subscribe_to: object = None, data: object = {}) -> object:

        rule = Rule(name, description)

        rule.set_filters(data.get(Const.FILTERS, []))
        rule.set_processing(data.get(Const.PROCESSING, []))
        rule.set_finally(data.get(Const.FINALLY, []))

        if isinstance(subscribe_to, str):
            subscribe_to = (subscribe_to,)
        for el in subscribe_to:
            event_router_factory().register(rule, el)



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections.abc import Mapping

from krules_core.route.router import DispatchPolicyConst

from krules_core.base_functions import RuleFunctionBase


## PAYLOAD FUNCTIONS ##############################################################

class UpdatePayload(RuleFunctionBase):
    """
    Update the payload _merging_ the received arguments interpreted as a dictionary
    """

    @staticmethod
    def _update(d, u):
        for k, v in u.items():
            if isinstance(v, Mapping):
                d[k] = UpdatePayload._update(d.get(k, {}), v)
            elif isinstance(v, list):
                v.extend(d.get(k, []))
                d[k] = v
            else:
                d[k] = v
        return d

    def execute(self, merge_dict):
        """
        Args:
            merge_dict: Dictionary to merge with
        """
        self._update(self.payload, merge_dict)


class SetPayloadProperties(RuleFunctionBase):
    """
    Set any number of properties in the payload. Existing properties are overridden
    """

    def execute(self, **kwargs):
        """
        Args:
              **kwargs: Each named paramenter is the key and the value to update with. The value can be a callable
                and it receives _eventually_ the previous value for that key. This means that the callable should be declared
                with a variable length number of arguments (*args)
        """
        for k, v in kwargs.items():
            if inspect.isfunction(v):
                args = []
                if k in self.payload:
                    args.append(self.payload[k])
                v = v(*args)
            self.payload[k] = v


class SetPayloadProperty(SetPayloadProperties):
    """
    Set a single property
    """

    def execute(self, property_name, value):
        """
        Args:
            property_name: A key in the dictionary,
            value: Value to set. Can be a callable (see SetPayloadProperties)
        """
        super().execute(**{property_name: value})


## SUBJECT FUNCTIONS ################################################################


class SetSubjectProperty(RuleFunctionBase):
    """
    Set a single property of the subject
    """
    def execute(self, property_name, value, extended=False, muted=False, cached=True):
        """
        Args:
            property_name: Name of the property to set. It may or may not exist
            value: Value to set. It can be a callable and receives (optionally) the current property value.
               If the property does not exist yet, it receives None
               (it is not possible to discriminate from an existing variable with the value None)
        """
        if extended:
            fn = lambda v: self.subject.set_ext(property_name, v, cached)
        else:
            fn = lambda v: self.subject.set(property_name, v, muted, cached)

        return fn(value)


class SetSubjectPropertySilently(SetSubjectProperty):
    """
    Set a property silently (no property changed event)
    """
    def execute(self, property_name, value, extended=False, cached=True, **kwargs):
        return super().execute(property_name, value, extended=extended, muted=True, cached=cached)


class StoreSubjectProperty(SetSubjectProperty):
    """
    Set a property directly to the storage without using the cache
    """
    def execute(self, property_name, value, extended=False, muted=False, **kwargs):
        return super().execute(property_name, value, extended=extended, muted=muted, cached=False)


class StoreSubjectPropertySilently(SetSubjectProperty):
    """
    Set a property directly to the storage without using the cache and without emit property changed events
    """
    def execute(self, property_name, value, extended=False, **kwargs):
        return super().execute(property_name, value, extended=extended, muted=True, cached=False)


class SetSubjectExtendedProperty(SetSubjectProperty):
    """
    Set an extended property of the subject
    """
    def execute(self, property_name, value, cached=True, **kwargs):
        return super().execute(property_name, value, extended=True, muted=True, cached=cached)


class SetSubjectProperties(RuleFunctionBase):
    """
        Set multiple properties in subject from dictionary. This is allowed only by using cache
        and not for extended properties
        Selectively emit property changed events
    """

    def execute(self, props, unmuted=[]):
        """
        Args:
            props: The properties to set
            unmuted: List of property names for which emit property changed events
        """
        for name, value in props.items():
            self.subject.set(name, value, muted=name not in unmuted)


class IncrementSubjectProperty(RuleFunctionBase):
    """
    Increment a numeric property in the subject. This is a conveniencs functions
    useful to accessing counters in a concurrent system. Implicitly
    call directly the storage backend bypassing the cache
    """
    def execute(self, property_name, amount=1, muted=False):
        if not isinstance(amount, (int, float, complex)):
            raise TypeError("amount must be a numeric type")
        return self.subject.set(
            property_name, lambda x: x is None and 0 + amount or x + amount, muted, cached=False)


class IncrementSubjectPropertySilently(IncrementSubjectProperty):

    def execute(self, property_name, amount=1, **kwargs):
        super().execute(property_name, amount, True)


class DecrementSubjectProperty(RuleFunctionBase):
    """
    Increment a numeric property in the subject. This is a conveniencs functions
    useful to accessing counters in a concurrent system. Implicitly
    call directly the storage backend bypassing the cache
    """
    def execute(self, property_name, amount=1, is_mute=False):
        if not isinstance(amount, (int, float, complex)):
            raise TypeError("amount must be a numeric type")

        return self.subject.set(
            property_name, lambda x: x is None and 0 - amount or x - amount, is_mute, cached=False)


class DecrementSubjectPropertySilently(DecrementSubjectProperty):

    def execute(self, property_name, amount=1, **kwargs):
        return super().execute(property_name, amount, True)


class StoreSubject(RuleFunctionBase):
    """
    Flush the cache
    """

    def execute(self):
        self.subject.store()


class FlushSubject(RuleFunctionBase):
    """
    Remove this subject and all of his properties
    """

    def execute(self):
        self.subject.flush()


#####################################################################################


class Route(RuleFunctionBase):
    """
    Produce an event inside and/or outside the ruleset according to _dispatch_policy_
    For "sending outside" the event we mean to deliver it to the dispatcher component
    """

    def execute(self, type=None, subject=None, payload=None, dispatch_policy=DispatchPolicyConst.DEFAULT):
        """
        Args:
            type: The event type. Default is current processing event type
            subject: New subject or the current subject as default
            payload: The payload of the event or the current payload
            dispatch_policy: Router -> dispatcher policy. Available choices are defined in
                 krules_core.route.router.DispatchPolicyConst as:

                    DEFAULT: Dispatched outside only when no handler is found in current ruleset
                    ALWAYS: Always dispatched even if an handler is found and processed in the current ruleset
                    NEVER: Never dispatched outside
                    DIRECT: Skip to search for a local handler and send outside directly
        """

        from krules_core.providers import event_router_factory
        if type is None:
            type = self.type
        if subject is None:
            subject = self.subject
        if payload is None:
            payload = self.payload

        event_router_factory().route(type, subject, payload, dispatch_policy=dispatch_policy)




class RaiseException(RuleFunctionBase):
    """
    Raise an exception
    """

    def execute(self, ex):

        raise ex




from .. import RuleConst


def get_value_from_payload_diffs(key, payload_diffs, default_value=None):
    for patch in payload_diffs:
        if patch["path"] == "/%s" % key and patch["op"] in ("add", "replace"):
            return patch["value"]
    return default_value


def check_operation_on_property(property_name, payload, operation):
    for patch in payload[RuleConst.PAYLOAD_DIFFS]:
        if patch["path"] == "/%s" % property_name and patch["op"] == operation:
            return True
    return False



from krules_core.subject import SubjectProperty, SubjectExtProperty

from krules_core.providers import subject_factory, subject_storage_factory
from dependency_injector import providers as providers
from krules_core.subject.empty_storage import EmptySubjectStorage


def setup_module(_):

    subject_storage_factory.override(
        providers.Factory(
            lambda *args, **kwargs:
                EmptySubjectStorage()
        )
    )


def teardown_module(_):
    subject_storage_factory.reset_last_overriding()


def test_load_store_and_flush():

    storage = subject_storage_factory("empty")

    storage.store(
        inserts=(
            SubjectProperty("p1", 1),
            SubjectProperty("p2", "2'3"),
            SubjectExtProperty("px1", 3),
            SubjectExtProperty("p1", "s1")
        )
    )

    props, ext_props = storage.load()

    assert len(props) == 0
    assert len(ext_props) == 0

    storage.store(
        updates=(
            SubjectProperty("p2", 3),
            SubjectExtProperty("p1", "s2"),
        ),
        deletes=(
            SubjectProperty("p1"),
            SubjectExtProperty("px1"),
        )
    )

    props, ext_props = storage.load()

    assert len(props) == 0
    assert len(ext_props) == 0

    storage.flush()

    assert len(props) == 0
    assert len(ext_props) == 0


def test_set_and_get():

    storage = subject_storage_factory("empty")
    storage.flush()

    assert storage.get("my_prop") is None

    new_value, old_value = storage.set(SubjectProperty("my_prop", 1))
    assert old_value is None
    assert new_value is None
    assert storage.get(SubjectProperty("my_prop")) is None

    storage.delete(SubjectProperty("my_prop"))
    new_value, old_value = storage.set(SubjectProperty("pset", 1), 0)
    assert old_value is None
    assert new_value is None
    new_value, old_value = storage.set(SubjectProperty("pset", "1'2"))
    assert new_value is None
    assert old_value is None
    assert storage.get(SubjectProperty("pset")) is None


def test_ext_props():

    storage = subject_storage_factory("test")

    storage.flush()

    storage.set(SubjectProperty("p1", 1))
    storage.set(SubjectProperty("p2", 2))
    storage.set(SubjectExtProperty("p3", 3))
    storage.set(SubjectExtProperty("p4", 4))

    props = storage.get_ext_props()

    assert len(props) == 0



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class RuleConst(object):
    """
    Basic consts
    """

    PROCESS_ID = "process_id"
    ORIGIN_ID = "origin_id"
    RULENAME = "name"
    DESCRIPTION = "description"
    SUBSCRIBE_TO = "subscribe_to"
    RULEDATA = "data"

    FILTERS = "filters"
    PROCESSING = "processing"
    FINALLY = "finally"

    TYPE = "type"
    SUBJECT = "subject"
    SECTION = "section"
    FUNC_NAME = "function"
    PAYLOAD = "payload"
    ARGS = "args"
    KWARGS = "kwargs"
    RETURNS = "returns"
    PROCESSED = "processed"
    EVENT_INFO = "event_info"

    GOT_ERRORS = "got_errors"
    EXCEPTION = "exception"
    EXC_INFO = "exc_info"
    EXC_EXTRA_INFO = "exc_extra_info"

    PAYLOAD_DIFFS = "payload_diffs"


class ConfigKeyConst(object):

    TYPE_TOPICS_PREFIX = "TYPE_TOPICS_PREFIX"
